#!/usr/bin/perl

use v5.12;     # or later to get "unicode_strings" feature
use strict;    # quote strings, declare variables
use warnings;  # on by default

binmode(STDOUT, ":utf8");
binmode(STDERR, ":utf8");

my $version = "1.01";

#use Data::Dumper;
use LWP::UserAgent;
use Getopt::Long;
use File::Basename;
#use Date::Calc qw(Delta_Days Decode_Month);
use Time::HiRes qw(usleep);
use Tie::File;
use Text::CSV;
use URI::Escape;
use XML::LibXML '1.70';
use if $^O eq 'MSWin32', 'Win32::Clipboard' => 'CF_UNICODETEXT';
use if $^O eq 'darwin', 'Mac::Pasteboard';

#my $sleep_microseconds = 500;
sub debug;
sub listclean;

my $prog = basename($0);

sub Usage {
    die @_ ? join('', @_, "\n") : '',
    <<ENDUSAGE, "\nStopped";
Usage: $prog <options>
    options:
    --configfile | -c <config file>		# read options from config file upon startup
    --debug      | -d				# enable debug output
    --escript    | -e				# executable perl code
    --efile      | -E <script file>		# executable perl code from specified file
    --field      | -f				# field(s) to change
    --help	 | -h				# output help and usage text
    --mpl					# output MPL
    --nomcws					# don't use mcws service
    --outfile    | -o <output file.mpl>		# use file specified as the MPL output file
    --password   | -P <username:password>	# MCWS username and password
    --server     | -S <server:port>		# MCWS server and port
    --testonly   | -t              		# test-only with mcws: no updates performed (implies -v)
    --verbose    | -v				# output operations more verbosely
ENDUSAGE
}

my @save_ARGV = @ARGV;
my %opts = ();
my %mcwsparams;

{
local $SIG{__WARN__} = sub { print "*** $_[0]\n" };

Getopt::Long::Configure('no_ignore_case');
GetOptions(\%opts,
	'configfile|c=s' => sub {  read_config_file($_[1]) },
	'debug|d',
	'escript|e=s',
	'efile|E=s',
	'field|f=s@',
	'help|h',
	'mpl',
	'nomcws',
	'outfile|o=s',
	'password|P=s'	 => sub { ($mcwsparams{'username'}, $mcwsparams{'password'}) = split(/:/, $_[1], 2) },
	'server|S=s'	 => sub { $mcwsparams{'server'} = $_[1] },
	'testonly|t',
	'verbose|v',
    ) or Usage;
}
debug "Command Line: @save_ARGV";

exists $opts{'help'} and Usage();
$opts{'verbose'} = 1 if exists $opts{'testonly'};
Usage "Use only one of -e or -E." if exists $opts{'efile'} and exists $opts{'escript'};
Usage "Use of -e or -E is required." if !exists $opts{'efile'} and !exists $opts{'escript'};

unless (exists $opts{'nomcws'}) {
    for (qw(server username password)) {
	exists $mcwsparams{$_} or die "Option --mcws used, but no $_ specified\nStopped"; 
    }
    mcws_do_mcc_command(\%mcwsparams, 21000);
}

# read the script from a file if necessary
if ($opts{'efile'}) {
    local $/ = undef;

    open EFILE, "<", $opts{'efile'} or
	die "Unable to open script file $opts{'efile'}. $!\nStopped";

    $opts{'escript'} = <EFILE>;
    close EFILE;
}


# Setup the default output file.
if ($^O eq 'MSWin32') {
    defined $opts{'outfile'} or
	$opts{'outfile'} = join('\\', $ENV{'HOMEPATH'}, 'Desktop', 'import.mpl');
}
else {
    defined $opts{'outfile'} or
	$opts{'outfile'} = join('/', $ENV{'HOME'}, 'Desktop', 'import.mpl');
}

my (@cliptracks, @clipheaders);

{
    my $clip;

    if ($^O eq 'MSWin32') {
	$clip = Win32::Clipboard();
	$_ = $clip->GetAs(&Win32::Clipboard::CF_UNICODETEXT);
	$_ = Encode::decode("UTF16-LE", $_);
    }
    else {
	use Encode qw(encode decode from_to FB_CROAK);

	$clip = Mac::Pasteboard->new ();
	# the Mac clipboard will sometimes be in UTF-16LE, other times in UTF-8
	my $c = $clip->paste('public.utf16-plain-text');
	my $decoded = eval { decode('UTF-16LE', $c, FB_CROAK) };
	$_ = decode('UTF-8', $decoded);
    }
}

# if there's a tab at the end of the line in the first row, we'll use the old format
if (/\t\R/o) {
    # before MC 19.0.36: work around broken TSV format (no quoting was done)
    my ($headers,$rows) = split /\R+/, $_, 2;
    # bogus: MC outputs trailing \t on the header row
    # also strip any extra line feeds, newlines
    $headers =~ s/[\t\013\010]*$//;
    @clipheaders = split /\t/, $headers;

    my $n = @clipheaders - 1;
    while ($rows =~ s#^([^\t]*(?:\t[^\t]*){$n})\R##oms) {
	my $r = $1;
	$r =~ s/\s*\R$//;
	debug "Row: '$r'\n";
	my @row = split(/\t/, $r);
	my %h = ();
	for (my $i = 0; $i < @clipheaders; $i++) {
	    $h{$clipheaders[$i]} = $row[$i];
	}
	push @cliptracks, \%h;
    }
}
else {
    # MC 19.0.36: use standard CSV format
    my $lref;
    my $sep_char = /\t/ ? "\t" : ',';
    my $csv = Text::CSV->new ( { binary => 1, sep_char => $sep_char } );
    $csv->eol("\015\015") if $^O eq 'darwin';
    open my $io, "<", \$_ or die $!;
    $lref = $csv->getline($io) or
	die "Clipboard does not appear to have copied track data.\nStopped";
    $csv->column_names ($lref);
    while (my $hr = $csv->getline_hr ($io)) {;
	push @cliptracks, $hr;
    }
    close $io;
}

@cliptracks > 0 or
    die "Clipboard does not appear to have copied track data.  Select the tracks and Edit->Copy.\nStopped";
grep /\bFilename\b/, keys %{$cliptracks[0]} or
    die "Clipboard does not have the Filename column, or clipboard data is not from MC.\nStopped";

if (my $outlist_ref = process_user_tracks(\@cliptracks)) {
    output_mpl($outlist_ref)	if $opts{'mpl'};
    send_to_mcws($outlist_ref)	if ! exists $opts{'nomcws'} and ! exists $opts{'testonly'};
}

# cycle through the user's tracks on the clipboard
sub process_user_tracks {
    my ($clipsref, $amgref) = @_;
    my @output_list = ();

    my $i = 1;
    for my $file (@$clipsref) {
	my %h = ();
	say "Filename: $file->{'Filename'}"  if $opts{'verbose'};
	for my $field (@{$opts{'field'}}) {
	    say "\told($field): '$file->{$field}'" if $opts{'verbose'};
	    $_ = $file->{$field};
	    {
		no strict;
		eval $opts{'escript'};
	    }
	    die "The expression:\n\n\t$opts{'escript'}\n\nfailed to execute: $@\nStopped" if $@;
	    say "\tnew($field): ", $_ eq $file->{$field} ? 'UNCHANGED' : $_   if $opts{'verbose'};
	    $h{$field} = $_    if $_ ne $file->{$field};
	}
	if (%h) {
	    $h{'Filename'} = $file->{'Filename'};
	    push @output_list, \%h;
	    debug 'Added to output list: : ', $file->{'Filename'};
	}
    }

    return \@output_list;
    exit;
}

# read and process a config file
sub read_config_file {
    my $configfile = shift;

    open my $io, "<", $configfile or
	die "can't open config file $configfile\n$!\nStopped";

    while (<$io>) {
	s/\R+$//;
	next if /^\s*$/;
	next if /^\s*#/;

	# Options:  command line options
	if (/Options:\s*(?<opts>.+)$/) {
	    unshift @ARGV, split /\s+/, $+{'opts'};
	    debug "Options: $+{'opts'}\n";
	}
	elsif (/MCWS_(server|username|password):\s*(.+?)\s*$/) {
	    $mcwsparams{$1} = $2;
	}
    }
    close $io;
}

sub open_file {
    my $filepath = shift;
    my $mpl;

    if ($filepath eq '-') {
	$mpl = \*STDOUT;
	delete $opts{'append'};		# ensure writing of the MPL preamble
	return $mpl;
    }
    else {
	my $openmode = '>';

	if (exists $opts{'append'}) {
	    if (-e $filepath and ! -z _) {
		my @lines;

		tie(@lines, 'Tie::File', $filepath) or
		    die "Can't update file $filepath\nStopped";
		delete $lines[-1] if $lines[-1] =~ /^<\/MPL>/;
		untie @lines;
		$openmode = '>>';
	    }
	    else {
		delete $opts{'append'};	# ensure writing of the MPL preamble
	    }
	}

	#my $encoding = ' :utf8';
	my $encoding = '';
	open($mpl, $openmode . $encoding, $filepath) or
	    die "Failed to open $filepath: $!\nStopped";
    }

    return $mpl;
}

sub listclean {
    local $_ = shift;

    s/\s*;$//;
    s/;/; /g;

    return $_;
}

sub output_mpl {
    my $lref = shift;
    return unless $lref;

    sub create_item {
	my $cellsref = shift;
	my $str;
	my $count = 0;

	for (keys %$$cellsref) {
	    if ($$cellsref->{$_} and $$cellsref->{$_} ne '' and $$cellsref->{$_} ne '-') {
		my $val = $$cellsref->{$_};
		$val =~ s/\&(?!amp;|quot;)/&amp;/g;
		$val =~ s/</&lt;/g;
		$val =~ s/>/&gt;/g;
		$str .= '<Field Name="' . $_ . '">'. $val . "</Field>\n";
		$count++;
	    }
	}
	unless ($count) {
	    say 'MPL: Nothing to update for file: ', $$cellsref->{'Filename'}   if $opts{'verbose'};
	    return '';
	}

	return "<Item>\n" . $str . "</Item>\n";
    }

    my $s;
    for (@$lref) {
	$s .= create_item(\$_);
    }

    if ($s) {
	my $mpl = open_file($opts{'outfile'});

	print $mpl qq/<?xml version="0.0" encoding="UTF-8" standalone="yes" ?>\n/,
		   qq/<MPL Version="2.0" Title="Created from $0 version $version">\n/;

	print $mpl $s, "</MPL>\n";
    }
}

sub send_to_mcws {
    my $lref = shift;
    return unless $lref;

    mcws_send_items(\%mcwsparams, \$_) for @$lref;
}

sub mcws_init {
    my $mcwsref = shift;

    $mcwsref->{'baseurl'} = join '', 'http://', $mcwsref->{'server'}, '/MCWS/v1';
    $mcwsref->{'ua'} = LWP::UserAgent->new(agent => 'Mozilla/5.0');
    $mcwsref->{'ua'}->timeout(10);
    $mcwsref->{'ua'}->env_proxy;
    $mcwsref->{'ua'}->cookie_jar({});
    $mcwsref->{'ua'}->credentials($mcwsref->{'server'}, 'Library Server', $mcwsref->{'username'} => $mcwsref->{'password'});
}

sub mcws_do_mcc_command {
    my $mcwsref = shift;

    mcws_init($mcwsref)  if ! exists $mcwsref->{'ua'};

    #http://localhost:52199/MCWS/v1/Control/MCC?Command=10000
    #http://localhost:52199/MCWS/v1/Control/MCC?Command=10005&Parameter=2
    my $url = join '/', $mcwsref->{'baseurl'}, 'Control/MCC?Command=' . $_[0] . (scalar @_ > 1 ? "&Parameter=$_[1]" : '');
    debug "Sending: $url";

    my $response = $mcwsref->{'ua'}->get($url);
    unless ($response->is_success) {
	die "MCWS write failed: \n$url\n", $response->status_line, "\n", $response->content;
    }
    $response->content =~ /<Response Status="OK"\/>/ or
	warn 'Unexpected server response: ', $response->content;

    sleep(1); # let MC settle
}

sub mcws_send_items {
    my ($mcwsref, $cellsref) = @_;

    my %params = (
	FileType	=> 'Filename',
	File		=> uri_escape_utf8($$cellsref->{'Filename'}),
    );

    mcws_init($mcwsref)  if ! exists $mcwsref->{'ua'};

    my (@fields, @values);
    debug "Filename: ", $params{'File'};
    while (my ($key, $value) = each %$$cellsref) {
	next if $key eq 'Filename';
	next if !defined $value or $value eq '' or $value eq '-';
	push @fields, $key;
	push @values, $value;
	debug "Added field($key): ", $key, " -> $value";
    }

    if (scalar @fields < 1) {
	say 'MCWS: Nothing to update for file: ', $params{'File'} if $opts{'verbose'};
	return;
    }

    if (scalar @fields == 1) {
	$params{'Field'} = uri_escape_utf8($fields[0]);
	$params{'Value'} = uri_escape_utf8($values[0]);
    }
    else {
	$params{'List'}	= 'CSV';

	my $csv = Text::CSV->new ( { binary => 1, sep_char => ',' } );

	# Fields
	$csv->combine(@fields) or die "Failed to combine field names into CSV\nStopped";
	$params{'Field'} = uri_escape_utf8($csv->string());

	# Values
	$csv->combine(@values) or die "Failed to combine field values into CSV\nStopped";
	$params{'Value'} = uri_escape_utf8($csv->string());
    }

    my $url = join '/', $mcwsref->{'baseurl'}, 'File/SetInfo?' . join '&', map { join('=', $_, $params{$_}) } keys %params;
    debug "Sending: $url";

    my $response = $mcwsref->{'ua'}->get($url);
    unless ($response->is_success) {
	die "MCWS write failed: \n$url\n", $response->status_line, "\n", $response->content;
    }
    $response->content =~ /<Response Status="OK"\/>/ or
	warn 'Unexpected server response: ', $response->content;
}

sub debug {
    return unless $opts{'debug'};

    say((split(/::/, (caller(1))[3] || 'main'))[-1], ': ', @_);
}

